from .generic import GenericAdapter
from ..utils import VAR_REF


def _replacer(match):
    gd = match.groupdict()
    if gd["dblquote"] is not None:
        return gd["dblquote"]
    elif gd["quote"] is not None:
        return gd["quote"]
    else:
        return f'{gd["lead"]}%({gd["var_name"]})s{gd["trail"]}'


class PyFormatAdapter(GenericAdapter):
    """Convert from named to pyformat parameter style"""

    def process_sql(self, _query_name, _op_type, sql):
        return VAR_REF.sub(_replacer, sql)



# use re2 if available.
# FIXME strange issue on debian buster with py3.7 apache wsgi flask… loading re2 "freezes" the process
try:
    import re2 as re
except ModuleNotFoundError:  # pragma: no cover
    import re  # type: ignore

VAR_REF = re.compile(
    r'(?P<dblquote>"[^"]+")|'
    r"(?P<quote>\'[^\']*\')|"
    r"(?P<lead>[^:]):(?P<var_name>[\w-]+)(?P<trail>[^:]?)"
)
"""Pattern to identifies colon-variables in SQL code"""


class SQLLoadException(Exception):
    """Raised when there is a problem loading SQL content from a file or directory"""

    pass


class SQLParseException(Exception):
    """Raised when there was a problem parsing the aiosql comment annotations in SQL"""

    pass



from .generic import GenericAdapter


class APSWAdapter(GenericAdapter):
    """
    APSW Adapter suitable for `named` parameter style and no with support.
    """

    def select(self, conn, _query_name, sql, parameters, record_class=None):
        cur = self._cursor(conn)
        try:
            cur.execute(sql, parameters)
            if record_class is None:
                results = cur.fetchall()
            else:
                # NOTE must use description on the fly
                results = [
                    record_class(**dict(zip([c[0] for c in cur.description], row))) for row in cur
                ]
        finally:
            cur.close()
        return results



from collections import defaultdict
from contextlib import asynccontextmanager

from ..utils import VAR_REF


class MaybeAcquire:
    def __init__(self, client, driver=None):
        self.client = client
        self._driver = driver

    async def __aenter__(self):
        if "acquire" in dir(self.client):
            self._managed_conn = await self.client.acquire()
            return self._managed_conn
        else:
            self._managed_conn = None
            return self.client

    async def __aexit__(self, exc_type, exc, tb):
        if self._managed_conn is not None:
            await self.client.release(self._managed_conn)


class AsyncPGAdapter:
    is_aio_driver = True

    def __init__(self):
        self.var_sorted = defaultdict(list)

    def process_sql(self, query_name, _op_type, sql):
        adj = 0

        for match in VAR_REF.finditer(sql):
            gd = match.groupdict()
            # Do nothing if the match is found within quotes.
            if gd["dblquote"] is not None or gd["quote"] is not None:
                continue

            var_name = gd["var_name"]
            if var_name in self.var_sorted[query_name]:
                replacement = f"${self.var_sorted[query_name].index(var_name) + 1}"
            else:
                replacement = f"${len(self.var_sorted[query_name]) + 1}"
                self.var_sorted[query_name].append(var_name)

            # Determine the offset of the start and end of the original
            # variable that we are replacing, taking into account an adjustment
            # factor based on previous replacements (see the note below).
            start = match.start() + len(gd["lead"]) + adj
            end = match.end() - len(gd["trail"]) + adj

            sql = sql[:start] + replacement + sql[end:]

            # If the replacement and original variable were different lengths,
            # then the offsets of subsequent matches will be wrong by the
            # difference.  Calculate an adjustment to apply to reconcile those
            # offsets with the modified string.
            #
            # The "- 1" is to account for the leading ":" character in the
            # original string.
            adj += len(replacement) - len(var_name) - 1

        return sql

    def maybe_order_params(self, query_name, parameters):
        if isinstance(parameters, dict):
            return [parameters[rk] for rk in self.var_sorted[query_name]]
        elif isinstance(parameters, tuple):
            return parameters
        else:
            raise ValueError(f"Parameters expected to be dict or tuple, received {parameters}")

    async def select(self, conn, query_name, sql, parameters, record_class=None):
        parameters = self.maybe_order_params(query_name, parameters)
        async with MaybeAcquire(conn) as connection:
            results = await connection.fetch(sql, *parameters)
            if record_class is not None:
                results = [record_class(**dict(rec)) for rec in results]
        return results

    async def select_one(self, conn, query_name, sql, parameters, record_class=None):
        parameters = self.maybe_order_params(query_name, parameters)
        async with MaybeAcquire(conn) as connection:
            result = await connection.fetchrow(sql, *parameters)
            if result is not None and record_class is not None:
                result = record_class(**dict(result))
        return result

    async def select_value(self, conn, query_name, sql, parameters):
        parameters = self.maybe_order_params(query_name, parameters)
        async with MaybeAcquire(conn) as connection:
            return await connection.fetchval(sql, *parameters)

    @asynccontextmanager
    async def select_cursor(self, conn, query_name, sql, parameters):
        parameters = self.maybe_order_params(query_name, parameters)
        async with MaybeAcquire(conn) as connection:
            stmt = await connection.prepare(sql)
            async with connection.transaction():
                yield stmt.cursor(*parameters)

    async def insert_returning(self, conn, query_name, sql, parameters):
        parameters = self.maybe_order_params(query_name, parameters)
        async with MaybeAcquire(conn) as connection:
            res = await connection.fetchrow(sql, *parameters)
            if res:
                return res[0] if len(res) == 1 else res
            else:
                return None

    async def insert_update_delete(self, conn, query_name, sql, parameters):
        parameters = self.maybe_order_params(query_name, parameters)
        async with MaybeAcquire(conn) as connection:
            await connection.execute(sql, *parameters)

    async def insert_update_delete_many(self, conn, query_name, sql, parameters):
        parameters = [self.maybe_order_params(query_name, params) for params in parameters]
        async with MaybeAcquire(conn) as connection:
            await connection.executemany(sql, parameters)

    async def execute_script(self, conn, sql):
        async with MaybeAcquire(conn) as connection:
            return await connection.execute(sql)



from .pyformat import PyFormatAdapter


class BrokenMySQLAdapter(PyFormatAdapter):
    """
    Work around PyMySQL and MySQLDB mishandling of empty parameters.

    See: https://github.com/PyMySQL/PyMySQL/issues/1059
    """

    def select(self, conn, name, sql, params, rc=None):
        return super().select(conn, name, sql, params or None, rc)

    def select_one(self, conn, name, sql, params, rc=None):
        return super().select_one(conn, name, sql, params or None, rc)

    def select_value(self, conn, name, sql, params):
        return super().select_value(conn, name, sql, params or None)

    def insert_update_delete(self, conn, name, sql, params):
        return super().insert_update_delete(conn, name, sql, params or None)

    # only called for mariadb, as mysql does not implement RETURNING
    def insert_returning(self, conn, name, sql, params):  # pragma: no cover
        return super().insert_returning(conn, name, sql, params or None)

    # left out with params: insert_update_delete_many, select_cursor



import sys
import inspect
from pathlib import Path
from types import MethodType
from typing import Any, Callable, List, Optional, Set, Tuple, cast

from .types import DriverAdapterProtocol, QueryDatum, QueryDataTree, QueryFn, SQLOperationType


def _params(args, kwargs):
    if len(kwargs) > 0:
        return kwargs
    else:
        return args


def _query_fn(
    fn: Callable[..., Any],
    name: str,
    doc: Optional[str],
    sql: str,
    operation: SQLOperationType,
    signature: Optional[inspect.Signature],
    floc: Optional[Tuple[Path, int]] = None,
) -> QueryFn:
    # TODO remove version workaround and pragmas when 3.7 support is dropped
    # FIXME should get the lineno as well?
    if floc and sys.version_info >= (3, 8, 0):  # pragma: no cover
        fname, lineno = floc
        fn.__code__ = fn.__code__.replace(co_filename=str(fname), co_firstlineno=lineno)  # type: ignore
    qfn = cast(QueryFn, fn)
    qfn.__name__ = name
    qfn.__doc__ = doc
    qfn.__signature__ = signature
    qfn.sql = sql
    qfn.operation = operation
    return qfn


# NOTE about coverage: because __code__ is set to reflect the actual SQL file
# source, coverage does note detect that the "fn" functions are actually called,
# hence the "no cover" hints.
def _make_sync_fn(query_datum: QueryDatum) -> QueryFn:
    query_name, doc_comments, operation_type, sql, record_class, signature, floc = query_datum
    if operation_type == SQLOperationType.INSERT_RETURNING:

        def fn(self: Queries, conn, *args, **kwargs):  # pragma: no cover
            return self.driver_adapter.insert_returning(
                conn, query_name, sql, _params(args, kwargs)
            )

    elif operation_type == SQLOperationType.INSERT_UPDATE_DELETE:

        def fn(self: Queries, conn, *args, **kwargs):  # pragma: no cover
            return self.driver_adapter.insert_update_delete(
                conn, query_name, sql, _params(args, kwargs)
            )

    elif operation_type == SQLOperationType.INSERT_UPDATE_DELETE_MANY:

        def fn(self: Queries, conn, *args, **kwargs):  # pragma: no cover
            return self.driver_adapter.insert_update_delete_many(
                conn, query_name, sql, *_params(args, kwargs)
            )

    elif operation_type == SQLOperationType.SCRIPT:

        def fn(self: Queries, conn, *args, **kwargs):  # pragma: no cover
            return self.driver_adapter.execute_script(conn, sql)

    elif operation_type == SQLOperationType.SELECT:

        def fn(self: Queries, conn, *args, **kwargs):  # pragma: no cover
            return self.driver_adapter.select(
                conn, query_name, sql, _params(args, kwargs), record_class
            )

    elif operation_type == SQLOperationType.SELECT_ONE:

        def fn(self: Queries, conn, *args, **kwargs):  # pragma: no cover
            return self.driver_adapter.select_one(
                conn, query_name, sql, _params(args, kwargs), record_class
            )

    elif operation_type == SQLOperationType.SELECT_VALUE:

        def fn(self: Queries, conn, *args, **kwargs):  # pragma: no cover
            return self.driver_adapter.select_value(conn, query_name, sql, _params(args, kwargs))

    else:
        raise ValueError(f"Unknown operation_type: {operation_type}")

    return _query_fn(fn, query_name, doc_comments, sql, operation_type, signature, floc)


def _make_async_fn(fn: QueryFn) -> QueryFn:
    async def afn(self: Queries, conn, *args, **kwargs):
        return await fn(self, conn, *args, **kwargs)

    return _query_fn(afn, fn.__name__, fn.__doc__, fn.sql, fn.operation, fn.__signature__)


def _make_ctx_mgr(fn: QueryFn) -> QueryFn:
    def ctx_mgr(self, conn, *args, **kwargs):
        return self.driver_adapter.select_cursor(conn, fn.__name__, fn.sql, _params(args, kwargs))

    return _query_fn(
        ctx_mgr, f"{fn.__name__}_cursor", fn.__doc__, fn.sql, fn.operation, fn.__signature__
    )


def _create_methods(query_datum: QueryDatum, is_aio: bool) -> List[QueryFn]:
    """Internal function to feed add_queries."""
    fn = _make_sync_fn(query_datum)
    if is_aio:
        fn = _make_async_fn(fn)

    ctx_mgr = _make_ctx_mgr(fn)

    if query_datum.operation_type == SQLOperationType.SELECT:
        return [fn, ctx_mgr]
    else:
        return [fn]


class Queries:
    """Container object with dynamic methods built from SQL queries.

    The `-- name` definition comments in the content of the SQL determine what the dynamic
    methods of this class will be named.

    **Parameters:**

    * **driver_adapter** - Either a string to designate one of the aiosql built-in database driver
    adapters. One of "sqlite3", "psycopg2", "aiosqlite", or "asyncpg". If you have defined your
    own adapter class, you can pass it's constructor.
    """

    def __init__(self, driver_adapter: DriverAdapterProtocol):
        self.driver_adapter: DriverAdapterProtocol = driver_adapter
        self.is_aio: bool = getattr(driver_adapter, "is_aio_driver", False)
        self._available_queries: Set[str] = set()

    @property
    def available_queries(self) -> List[str]:
        """Returns listing of all the available query methods loaded in this class.

        **Returns:** `List[str]` List of dot-separated method accessor names.
        """
        return sorted(self._available_queries)

    def __repr__(self):
        return "Queries(" + self.available_queries.__repr__() + ")"

    def add_query(self, query_name: str, fn: Callable):
        """Adds a new dynamic method to this class.

        **Parameters:**

        * **query_name** - The method name as found in the SQL content.
        * **fn** - The loaded query function.
        """
        setattr(self, query_name, fn)
        self._available_queries.add(query_name)

    def add_queries(self, queries: List[QueryFn]):
        """Add query methods to `Queries` instance."""
        for fn in queries:
            query_name = fn.__name__.rpartition(".")[2]
            self.add_query(query_name, MethodType(fn, self))

    def add_child_queries(self, child_name: str, child_queries: "Queries"):
        """Adds a Queries object as a property.

        **Parameters:**

        * **child_name** - The property name to group the child queries under.
        * **child_queries** - Queries instance to add as sub-queries.
        """
        setattr(self, child_name, child_queries)
        for child_query_name in child_queries.available_queries:
            self._available_queries.add(f"{child_name}.{child_query_name}")

    def load_from_list(self, query_data: List[QueryDatum]):
        """Load Queries from a list of `QuaryDatum`"""
        for query_datum in query_data:
            self.add_queries(_create_methods(query_datum, self.is_aio))
        return self

    def load_from_tree(self, query_data_tree: QueryDataTree):
        """Load Queries from a `QuaryDataTree`"""
        for key, value in query_data_tree.items():
            if isinstance(value, dict):
                self.add_child_queries(key, Queries(self.driver_adapter).load_from_tree(value))
            else:
                self.add_queries(_create_methods(value, self.is_aio))
        return self



from contextlib import contextmanager


class GenericAdapter:
    """
    Generic AioSQL Adapter suitable for `named` parameter style and no with support.
    """

    def __init__(self, driver=None):
        self._driver = driver

    def process_sql(self, _query_name, _op_type, sql):
        """Preprocess SQL query."""
        return sql

    def _cursor(self, conn):
        """Get a cursor from a connection."""
        return conn.cursor()

    def select(self, conn, _query_name, sql, parameters, record_class=None):
        cur = self._cursor(conn)
        try:
            cur.execute(sql, parameters)
            results = cur.fetchall()
            if record_class is not None and len(results) > 0:
                column_names = [c[0] for c in cur.description]
                results = [record_class(**dict(zip(column_names, row))) for row in results]
        finally:
            cur.close()
        return results

    def select_one(self, conn, _query_name, sql, parameters, record_class=None):
        cur = self._cursor(conn)
        try:
            cur.execute(sql, parameters)
            result = cur.fetchone()
            if result is not None and record_class is not None:
                column_names = [c[0] for c in cur.description]
                result = record_class(**dict(zip(column_names, result)))
        finally:
            cur.close()
        return result

    def select_value(self, conn, _query_name, sql, parameters):
        cur = self._cursor(conn)
        try:
            cur.execute(sql, parameters)
            result = cur.fetchone()
        finally:
            cur.close()
        return result[0] if result else None

    @contextmanager
    def select_cursor(self, conn, _query_name, sql, parameters):
        cur = self._cursor(conn)
        cur.execute(sql, parameters)
        try:
            yield cur
        finally:
            cur.close()

    def insert_update_delete(self, conn, _query_name, sql, parameters):
        cur = self._cursor(conn)
        cur.execute(sql, parameters)
        rc = cur.rowcount if hasattr(cur, "rowcount") else -1
        cur.close()
        return rc

    def insert_update_delete_many(self, conn, _query_name, sql, parameters):
        cur = self._cursor(conn)
        cur.executemany(sql, parameters)
        rc = cur.rowcount if hasattr(cur, "rowcount") else -1
        cur.close()
        return rc

    def insert_returning(self, conn, _query_name, sql, parameters):
        # very similar to select_one but the returned value
        cur = self._cursor(conn)
        cur.execute(sql, parameters)
        res = cur.fetchone()
        cur.close()
        return res[0] if res and len(res) == 1 else res

    def execute_script(self, conn, sql):
        cur = self._cursor(conn)
        cur.execute(sql)
        msg = cur.statusmessage if hasattr(cur, "statusmessage") else "DONE"
        cur.close()
        return msg



from .aiosql import from_path, from_str, register_adapter
from .utils import SQLParseException, SQLLoadException

# TODO python >= 3.8
#   from importlib.metadata import version
#   __version__ = version("aiosql")
import pkg_resources as pkg  # type: ignore

__version__ = pkg.require("aiosql")[0].version

__all__ = ["from_path", "from_str", "register_adapter", "SQLParseException", "SQLLoadException"]



# standard adapters
from .pyformat import PyFormatAdapter
from .generic import GenericAdapter
from .sqlite3 import SQLite3Adapter

# async adapters
from .aiosqlite import AioSQLiteAdapter
from .asyncpg import AsyncPGAdapter

# silence flake8 F401 warning:
_ALL = [
    PyFormatAdapter,
    GenericAdapter,
    SQLite3Adapter,
    AioSQLiteAdapter,
    AsyncPGAdapter,
]



from .generic import GenericAdapter


class SQLite3Adapter(GenericAdapter):
    # overwrites some methods using sqlite3-specific non-standard methods

    def insert_returning(self, conn, _query_name, sql, parameters):
        cur = self._cursor(conn)
        try:
            cur.execute(sql, parameters)
            results = cur.lastrowid
        finally:
            cur.close()
        return results

    def execute_script(self, conn, sql):
        conn.executescript(sql)
        return "DONE"
